#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#define STRING_SIZE 300
#define NCHANNEL 4

static FILE *file_channel[NCHANNEL] = {NULL, NULL, NULL, NULL};
static int *template_grid_size[NCHANNEL];
static long int *template_byte_loc[NCHANNEL];
static int n_template[NCHANNEL];
static int rw_channel[NCHANNEL];


static void trim_fort(char *cstring, int clen, char *fstring, int flen);

/*
 * call by fortran: to open the grib channel for writes
 *
 * call grb_wopen(ichannel, grbout, template, grid_size, ntemplate, compress_type, options, ierr)
 *
 * INPUT
 *
 * integer ichannel                       an integer from 0..NCHANNEL  (currently 3)
 * character*(*) grbout                   grib output filename
 * character*(*) template                 grib template
 * integer grid_size(ntemplate)	  	  grid size for each template
 * integer ntemplate			  number of templates
 * character*(*) compress_type            c1, c2, c3, j, etc
 * character*(*) options                  can set to "", see documentation for more details
 *
 * OUTPUT
 *
 * integer ierr                           error code
 *
 *
 * call by fortran: to write a grid
 *
 * call grb_write(ichannel, itemplate, data, metadata, ierr)
 *
 * INPUT
 *
 * integer ichannel                       an integer from 0..NCHANNEL  (currently 3)
 * integer itemplate                      an integer with template number (grib message number)
 * real data(ndata)                       ndata is defined in grb_wropen)
 * character*(*) metadata                 see documentation
 *
 * OUTPUT
 *
 * integer ierr                           error code
 *   1 = bad channel number
 *   2 = channel is not open
 *
 *
 * call by fortran: to close the grib channel
 *
 * call grb_close_(ichannel, ierr) {
 *
 * OUTPUT
 *
 * integer ierr                           error code
 *   1 = bad channel number
 *   2 = channel is not open
 *
 * v0.9    - 7/26/2013 test Wesley Ebisuzaki
 * v0.91   - 7/26/2013 test Wesley Ebisuzaki
 * v0.92   - 7/26/2013 test Wesley Ebisuzaki, scaling works
 * v0.93     8/16/2013 multiple template support, more error msgs  Wesley Ebisuzaki,
 */

void grb_wopen_(int *channel, char *out_file, char *template_file, int *grid_size, int *ntemplate, 
   char *compress_type, char *options, int *err, 
   int outlen, int templatelen, int compresslen, int optionslen) {

    char wgrib2_cmd[STRING_SIZE];
    char out[STRING_SIZE];
    char template[STRING_SIZE];
    char compress[STRING_SIZE];
    char option[STRING_SIZE];
    FILE *in;
    int i, j, k, n, ch;
    long int loc;

    /* convert fortran strings to C strings */
    trim_fort(out, STRING_SIZE, out_file, outlen);
    trim_fort(template, STRING_SIZE, template_file, templatelen);
    trim_fort(compress, STRING_SIZE, compress_type, compresslen);
    trim_fort(option, STRING_SIZE, options, optionslen);

    ch = *channel;
    n = *ntemplate;

    if (sizeof(int) != 4) {
        fprintf(stderr,"ERROR grb_wopen: size of int ne 4, compile problem\n");
        *err = 1;
        return;
    }

    if (ch < 0) {
        fprintf(stderr,"ERROR grb_wopen: negative channel = %d\n", ch);
        *err = 2;
        return;
    }
    if (ch >= NCHANNEL) {
        fprintf(stderr,"ERROR grb_wopen: channel too large %d >= %d\n", n, NCHANNEL);
        *err = 3;
        return;
    }

    if (n <= 0) {
        fprintf(stderr,"ERROR grb_wopen: bad ntemplate = %d\n", n);
        *err = 4;
        return;
    }
    if (n > 2000) {
        fprintf(stderr,"ERROR grb_wopen: ntemplate too big %d\n", n);
        *err = 5;
        return;
    }

    template_grid_size[ch] = malloc( n * sizeof(int));
    template_byte_loc[ch] = malloc( n * sizeof(long int));

    if (template_grid_size[ch] == NULL || template_byte_loc[ch] == NULL) {
        fprintf(stderr,"ERROR grb_wopen: memory allocatin problem\n");
        *err = 6;
        return;
    }
    n_template[ch] = 0;

    sprintf(wgrib2_cmd, "wgrib2 %s -for 1:%d -npts", template, n);
    in = popen(wgrib2_cmd, "r");
    if (in == NULL) {
        fprintf(stderr,"ERROR grb_wopen: problem executing %s\n", wgrib2_cmd);
        *err = 7;
        return;
    }

    for (i = 0; i < n; i++) {
        if (fscanf(in,"%d:%ld:npts=%d", &j,&loc,&k) != 3) {
            fprintf(stderr,"ERROR grb_wopen: could not find grib2 message\n");
            *err = 8;
            return;
	}

	if (j != i+1) {
            fprintf(stderr,"ERROR grb_wopen: expecting msg %d found %d\n", i, j);
            *err = 9;
            return;
	}
	if (k != grid_size[i]) {
            fprintf(stderr,"ERROR grb_wopen: grid size mismatch template %d grid_size req %d found %d\n",
		   i+1, grid_size[i], k);
            *err = 10;
            return;
	}
	template_grid_size[ch][i] = k;
	template_byte_loc[ch][i] = loc;
    }
    n_template[ch] = n;

    pclose(in);

    sprintf(wgrib2_cmd, "wgrib2 -i %s -inv /dev/null -set_grib_type %s -import_bin - %s -set_metadata - -grib_out %s", 
       template, compress, option, out);
    printf("grb channel(%d):%s\n", *channel, wgrib2_cmd);
    file_channel[*channel]  = popen(wgrib2_cmd,"w");
    rw_channel[*channel] = 0;
    *err = 0;
    return;
}


void grb_write_(int *channel, int *template, float *data, char *metadata, int *ierr, int metadatalen) {
    int i, j, ch, templateno,  ndata;
    char meta[STRING_SIZE];
    FILE *out;

    /* convert fortran strings to C strings */
    trim_fort(meta, STRING_SIZE, metadata, metadatalen);

    ch = *channel;
    templateno = *template - 1;

    if (ch < 0 || ch >= NCHANNEL) {
        fprintf(stderr,"ERROR grb_write: bad channel number %d\n", ch);
        *ierr = 1;
        return;
    }
    if (templateno < 0 || templateno >= n_template[ch]) {
        fprintf(stderr,"ERROR grb_write: bad template number %d (max %d) channel %d\n", templateno, 
		n_template[ch], ch);
        *ierr = 2;
        return;
    }

    out = file_channel[ch];
    if (out == NULL) {
        fprintf(stderr,"ERROR grb_write: no open file\n");
	*ierr = 3;
	return;
    }

    ndata = template_grid_size[ch][templateno];

    *ierr = 0;

    if (fprintf(out,"%d:%ld\n",templateno+1, template_byte_loc[ch][templateno]) < 0) {
	*ierr = 4;
	return;
    }

    /* write f77 header */
    i = ndata * 4;
    j = fwrite(&i, 4, 1, out);
    if (j != 1) { *ierr = 5; return; }

    j = fwrite(data, 4, ndata, out);
    if (j != ndata) { *ierr = 6; return; }

    /* write f77 trailer */
    j = fwrite(&i, 4, 1, out);
    if (j != 1) { *ierr = 7; return; }

    if (fprintf(out,"1:0:%s\n",meta) < 0) {
	*ierr = 4;
	return;
    }

    fflush(out);
    return;
}

void grb_close_(int *channel, int *ierr) {
    *ierr = 0;
    return;

    int ichannel;

    ichannel = *channel;

    if (ichannel < 0 || ichannel > NCHANNEL) {
        *ierr = 1;
        return;
    }

    if (file_channel[ichannel] == NULL) {
        *ierr = 2;
        return;
    }

    *ierr = 0;

    fflush(file_channel[ichannel]);
    pclose(file_channel[ichannel]);
    file_channel[ichannel] = NULL;
    free(template_grid_size[ichannel]);
    free(template_byte_loc[ichannel]);
    n_template[ichannel] = 0;
    return;
}

/* convert a fortran string to a C string with trim of trailing blanks */

static void trim_fort(char *cstring, int clen, char *fstring, int flen) {
    int len, i;

    len = clen-1 < flen ? clen-1 : flen;
    memcpy(cstring, fstring, len);
    cstring[len] = 0;

    i = strlen(cstring);
    while (i && cstring[i-1] == ' ') {
        cstring[--i] = 0;
    }

    return;
}
